/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.common.binaryfile;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.OutputStream;
import java.util.Arrays;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.core.task.SimpleAsyncTaskExecutor;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.errorhandling.TechnicalException;
import io.grpc.stub.CallStreamObserver;
import lombok.SneakyThrows;

class GrpcBinaryFileServerDownloaderITCase {

	@Mock
	private CallStreamObserver<GrpcResponseDummy> callObserver;
	private SimpleAsyncTaskExecutor taskExecutor;
	private GrpcBinaryFileServerDownloader<GrpcResponseDummy> downloader;

	@BeforeEach
	void init() {
		taskExecutor = new SimpleAsyncTaskExecutor();
		when(callObserver.isReady()).thenReturn(true);
	}

	@AfterEach
	void cleanup() {
		taskExecutor.close();
	}

	@Nested
	class OnNoError {

		private static final int DOWNLOAD_DATA_LENGTH = (int) (GrpcBinaryFileServerDownloader.CHUNK_SIZE * 1.5);
		@Captor
		private ArgumentCaptor<GrpcResponseDummy> captor;

		@BeforeEach
		void init() {
			downloader = spy(downloaderBuilder()
					.downloadConsumer(this::downloadData)
					.build());
		}

		@SneakyThrows
		@Test
		void shouldCallOnCompleted() {
			start();

			verify(callObserver).onCompleted();
		}

		@Test
		void shouldReadAllData() {
			start();

			verify(callObserver, times(2)).onNext(captor.capture());
			var totalBytesRead = captor.getAllValues().stream().mapToInt(GrpcResponseDummy::bytesRead).sum();
			assertThat(totalBytesRead).isEqualTo(DOWNLOAD_DATA_LENGTH);
		}

		@Test
		void shouldCloseStreams() {
			start();

			verify(downloader).closeInputStream();
			verify(downloader).closeOutputStream();
		}

		@Test
		void shouldCompleteIfDownloadConsumerClosedOutputStream() {
			downloader = spy(downloaderBuilder()
					.downloadConsumer(this::downloadDataAndCloseStream)
					.build());

			start();

			verify(callObserver).onCompleted();
		}

		@SneakyThrows
		private void downloadData(OutputStream outputStream) {
			byte[] bytes = new byte[DOWNLOAD_DATA_LENGTH];
			Arrays.fill(bytes, (byte) 1);
			outputStream.write(bytes);
		}

		@SneakyThrows
		private void downloadDataAndCloseStream(OutputStream outputStream) {
			downloadData(outputStream);
			outputStream.close();
			Thread.sleep(100); // delay, so that the onReadyHandler gets end of stream before this method returns
		}
	}

	@Nested
	class OnError {

		private final Throwable error = new TechnicalException("error");

		@BeforeEach
		void init() {
			downloader = spy(downloaderBuilder()
					.downloadConsumer(this::downloadData)
					.build());
		}

		@Test
		void shouldThrowException() {
			assertThatThrownBy(GrpcBinaryFileServerDownloaderITCase.this::start).isInstanceOf(TechnicalException.class);
		}

		@Test
		void shouldNotCallOnCompleted() {
			catchException(GrpcBinaryFileServerDownloaderITCase.this::start);

			verify(callObserver, never()).onCompleted();
		}

		@Test
		void shouldCloseStreams() {
			catchException(GrpcBinaryFileServerDownloaderITCase.this::start);

			verify(downloader).closeInputStream();
			verify(downloader).closeOutputStream();
		}

		@SneakyThrows
		private void downloadData(OutputStream outputStream) {
			throw error;
		}
	}

	private GrpcBinaryFileServerDownloader.GrpcBinaryFileServerDownloaderBuilder<GrpcResponseDummy> downloaderBuilder() {
		return GrpcBinaryFileServerDownloader.<GrpcResponseDummy>builder()
				.taskExecutor(taskExecutor)
				.callObserver(callObserver)
				.chunkBuilder(this::buildChunk);
	}

	private GrpcResponseDummy buildChunk(ByteString data) {
		return new GrpcResponseDummy(data.size());
	}

	private void start() {
		downloader.start();
		downloader.sendChunks();
	}

	private record GrpcResponseDummy(int bytesRead) {}
}
